# Copyright 2024-2025 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for the LSO package."""

from concurrent.futures import ThreadPoolExecutor

from lso.config import settings

_executor = None


def get_thread_pool() -> ThreadPoolExecutor:
    """Initialize or return a cached ThreadPoolExecutor for local asynchronous execution."""
    global _executor  # noqa: PLW0603
    if _executor is None:
        _executor = ThreadPoolExecutor(max_workers=settings.MAX_THREAD_POOL_WORKERS)

    return _executor



# Copyright 2024-2025 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module defines tasks for executing Ansible playbooks asynchronously using Celery.

The primary task, `run_playbook_proc_task`, runs an Ansible playbook and sends a POST request with
the results to a specified callback URL.
"""

import logging
from collections.abc import Callable
from typing import Any
from uuid import UUID

import requests
from ansible_runner import Runner, run
from starlette import status

from lso.config import settings
from lso.schema import ExecutableRunResponse
from lso.worker import RUN_EXECUTABLE, RUN_PLAYBOOK, celery

logger = logging.getLogger(__name__)


class CallbackFailedError(Exception):
    """Exception raised when a callback url can't be reached."""


def playbook_event_handler_factory(
    progress: str | None, *, progress_is_incremental: bool
) -> Callable[[dict], bool] | None:
    """Create an event handler for Ansible playbook runs.

    This is used to send incremental progress updates to the external system that called for this playbook to be run.

    :param str progress: The progress URL where the external system expects to receive updates.
    :param bool progress_is_incremental: Whether progress updates are sent incrementally, or contain the whole history
                                         of event data.
    """
    events_stdout = []

    def _playbook_event_handler(event: dict) -> bool:
        event_data = event["stdout"].strip()
        if not event_data:
            return False

        event_data_lines = event_data.split("\r\n")
        if progress_is_incremental:
            emit_body = event_data_lines
        else:
            events_stdout.extend(event_data_lines)
            emit_body = events_stdout

        requests.post(str(progress), json={"progress": emit_body}, timeout=settings.REQUEST_TIMEOUT_SEC)
        return True

    if progress:
        return _playbook_event_handler
    return None


def playbook_finished_handler_factory(callback: str | None, job_id: str) -> Callable[[Runner], None] | None:
    """Create an event handler for finished Ansible playbook runs.

    Once Ansible runner is finished, it will call the handler method created by this factory before teardown.

    :param str callback: The callback URL that ansible runner should report to.
    :param str job_id: The job ID of this playbook run, used for reporting.
    :return Callable: A handler method that sends one request to the callback URL.
    """

    def _playbook_finished_handler(runner: Runner) -> None:
        playbook_output = runner.stdout.read().split("\n")
        playbook_output = [line for line in playbook_output if line.strip()]

        payload = {
            "status": runner.status,
            "job_id": job_id,
            "output": playbook_output,
            "return_code": int(runner.rc),
        }

        response = requests.post(str(callback), json=payload, timeout=settings.REQUEST_TIMEOUT_SEC)
        if not (status.HTTP_200_OK <= response.status_code < status.HTTP_300_MULTIPLE_CHOICES):
            msg = f"Callback failed: {response.text}, url: {callback}"
            raise CallbackFailedError(msg)

    if callback:
        return _playbook_finished_handler
    return None


@celery.task(name=RUN_PLAYBOOK)  # type: ignore[misc]
def run_playbook_proc_task(
    job_id: str,
    playbook_path: str,
    extra_vars: dict[str, Any],
    inventory: dict[str, Any] | str,
    callback: str | None,
    progress: str | None,
    *,
    progress_is_incremental: bool,
) -> None:
    """Celery task to run a playbook.

    :param str job_id: Identifier of the job being executed.
    :param str playbook_path: Path to the playbook to be executed.
    :param dict[str, Any] extra_vars: Extra variables to pass to the playbook.
    :param dict[str, Any] | str inventory: Inventory to run the playbook against.
    :param str callback: Callback URL for status updates.
    :param str progress: URL for sending progress updates.
    :param bool progress_is_incremental: Whether progress updates include all past progress.
    :return: None
    """
    msg = f"playbook_path: {playbook_path}, callback: {callback}"
    logger.info(msg)
    run(
        playbook=playbook_path,
        inventory=inventory,
        extravars=extra_vars,
        event_handler=playbook_event_handler_factory(progress, progress_is_incremental=progress_is_incremental),
        finished_callback=playbook_finished_handler_factory(callback, job_id),
    )


@celery.task(name=RUN_EXECUTABLE)  # type: ignore[misc]
def run_executable_proc_task(job_id: str, executable_path: str, args: list[str], callback: str | None) -> None:
    """Celery task to run an arbitrary executable and notify via callback.

    Executes the executable with the provided arguments and posts back the result if a callback URL is provided.
    """
    from lso.execute import run_executable_sync  # noqa: PLC0415

    msg = f"Executing executable: {executable_path} with args: {args}, callback: {callback}"
    logger.info(msg)
    result = run_executable_sync(executable_path, args)

    if callback:
        payload = ExecutableRunResponse(
            job_id=UUID(job_id),
            result=result,
        ).model_dump(mode="json")

        def _raise_callback_error(message: str, error: Exception | None = None) -> None:
            if error:
                raise CallbackFailedError(message) from error
            raise CallbackFailedError(message)

        try:
            response = requests.post(str(callback), json=payload, timeout=settings.REQUEST_TIMEOUT_SEC)
            if not (status.HTTP_200_OK <= response.status_code < status.HTTP_300_MULTIPLE_CHOICES):
                msg = f"Callback failed: {response.text}, url: {callback}"
                _raise_callback_error(msg)
        except Exception as e:
            error_msg = f"Callback error: {e}"
            logger.exception(error_msg)
            _raise_callback_error(error_msg, e)



# Copyright 2023-2025 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for loading and managing configuration settings for the LSO app.

Uses `pydantic`'s `BaseSettings` to load settings from environment variables.
"""

import os
from enum import Enum

from pydantic_settings import BaseSettings


class ExecutorType(Enum):
    """Enum representing the types of executors available for task execution."""

    WORKER = "celery"
    THREADPOOL = "threadpool"


class Config(BaseSettings):
    """The set of parameters required for running :term:`LSO`."""

    TESTING: bool = False
    ANSIBLE_PLAYBOOKS_ROOT_DIR: str = "/path/to/ansible/playbooks"
    EXECUTABLES_ROOT_DIR: str = "/path/to/executables"
    EXECUTOR: ExecutorType = ExecutorType.THREADPOOL
    MAX_THREAD_POOL_WORKERS: int = min(32, (os.cpu_count() or 1) + 4)
    REQUEST_TIMEOUT_SEC: int = 10
    CELERY_BROKER_URL: str = "redis://localhost:6379/0"
    CELERY_RESULT_BACKEND: str = "redis://localhost:6379/0"
    CELERY_RESULT_EXPIRES: int = 3600
    WORKER_QUEUE_NAME: str | None = None
    EXECUTABLE_TIMEOUT_SEC: int = 300


settings = Config()



# Copyright 2023-2025 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Environment module for setting up logging."""

import json
import logging.config
import os
from pathlib import Path

LOGGING_DEFAULT_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {"simple": {"format": "%(asctime)s - %(name)s (%(lineno)d) - %(levelname)s - %(message)s"}},
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "stream": "ext://sys.stdout",
        }
    },
    "loggers": {"resource_management": {"level": "DEBUG", "handlers": ["console"], "propagate": False}},
    "root": {"level": "INFO", "handlers": ["console"]},
}


def setup_logging() -> None:
    """Set up logging using the configured filename.

    If ``LOGGING_CONFIG`` is defined in the environment, use this for the filename, otherwise use
    ``LOGGING_DEFAULT_CONFIG``.
    """
    logging_config = LOGGING_DEFAULT_CONFIG
    if "LOGGING_CONFIG" in os.environ:
        filename = os.environ["LOGGING_CONFIG"]
        config_file = Path(filename).read_text()
        logging_config = json.loads(config_file)

    logging.config.dictConfig(logging_config)



# Copyright 2023-2025 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default route located at the root URL /.

For now only includes a single endpoint that responds with the current version of the API and LSO.
"""

from importlib import metadata

from fastapi import APIRouter
from pydantic import BaseModel, constr

API_VERSION = "1.0"
VersionString = constr(pattern=r"\d+\.\d+")

router = APIRouter()


class Version(BaseModel):
    """Simple model for returning a version number of both the API and the `lso` module."""

    api: VersionString  # type: ignore[valid-type]
    module: VersionString  # type: ignore[valid-type]


@router.get("/version")
def version() -> Version:
    """Return the version numbers of the API version, and the module version.

    :return: Version object with both API and `lso` versions numbers.
    """
    return Version(api=API_VERSION, module=metadata.version("orchestrator-lso"))



# Copyright 2023-2025 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that gathers common API responses and data models."""

from pathlib import Path
from typing import Any
from uuid import UUID, uuid4

from pydantic import HttpUrl

from lso.config import ExecutorType, settings
from lso.tasks import run_playbook_proc_task
from lso.utils import get_thread_pool


def get_playbook_path(playbook_name: Path) -> Path:
    """Get the path of a playbook on the local filesystem."""
    return Path(settings.ANSIBLE_PLAYBOOKS_ROOT_DIR) / playbook_name


def run_playbook(
    playbook_path: Path,
    extra_vars: dict[str, Any],
    inventory: dict[str, Any] | str,
    callback: HttpUrl | None,
    progress: HttpUrl | None,
    *,
    progress_is_incremental: bool,
) -> UUID:
    """Run an Ansible playbook against a specified inventory.

    :param Path playbook_path: Playbook to be executed.
    :param dict[str, Any] extra_vars: Any extra vars needed for the playbook to run.
    :param dict[str, Any] | str inventory: The inventory that the playbook is executed against.
    :param HttpUrl callback: Callback URL where the playbook should send a status update when execution is completed.
                             This is used for workflow-orchestrator to continue with the next step in a workflow.
    :return UUID: Job ID of the launched playbook.
    """
    job_id = uuid4()
    callback_str = None
    progress_str = None
    if callback:
        callback_str = str(callback)
    if progress:
        progress_str = str(progress)

    if settings.EXECUTOR == ExecutorType.THREADPOOL:
        executor = get_thread_pool()
        executor_handle = executor.submit(
            run_playbook_proc_task,
            str(job_id),
            str(playbook_path),
            extra_vars,
            inventory,
            callback_str,
            progress_str,
            progress_is_incremental=progress_is_incremental,
        )
        if settings.TESTING:
            executor_handle.result()

    elif settings.EXECUTOR == ExecutorType.WORKER:
        run_playbook_proc_task.delay(
            str(job_id),
            str(playbook_path),
            extra_vars,
            inventory,
            callback_str,
            progress_str,
            progress_is_incremental=progress_is_incremental,
        )

    return job_id



# Copyright 2024-2025 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for defining the schema for running arbitrary executables."""

from enum import StrEnum
from uuid import UUID

from pydantic import BaseModel, model_validator


class JobStatus(StrEnum):
    """Enumeration of possible job statuses."""

    SUCCESSFUL = "successful"
    FAILED = "failed"


class ExecutionResult(BaseModel):
    """Model for capturing the result of an executable run."""

    output: str
    return_code: int
    status: JobStatus

    @model_validator(mode="before")
    def populate_status(cls, values: dict) -> dict:
        """Set the status based on the return code."""
        rc = values.get("return_code")
        if rc is not None:
            values["status"] = JobStatus.SUCCESSFUL if rc == 0 else JobStatus.FAILED

        return values


class ExecutableRunResponse(BaseModel):
    """Response for running an arbitrary executable."""

    job_id: UUID
    result: ExecutionResult | None = None



# Copyright 2023-2025 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FastAPI route for running arbitrary executables."""

import asyncio
from pathlib import Path
from typing import Annotated
from uuid import uuid4

from fastapi import APIRouter, HTTPException, status
from pydantic import AfterValidator, BaseModel, HttpUrl

from lso.execute import get_executable_path, run_executable_async, run_executable_sync
from lso.schema import ExecutableRunResponse

router = APIRouter()


def _executable_path_validator(executable_name: Path) -> Path:
    """Validate that the executable exists, is a file, and is marked as executable."""
    executable_path = get_executable_path(executable_name)
    if not executable_path.exists():
        msg = f"Executable '{executable_path}' does not exist."
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=msg)
    if not executable_path.is_file():
        msg = f"Executable '{executable_name}' is not a valid file."
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=msg)
    if not executable_path.stat().st_mode & 0o111:
        msg = f"Executable '{executable_name}' is not marked as executable."
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail=msg)
    return executable_path


ExecutableName = Annotated[Path, AfterValidator(_executable_path_validator)]


class ExecutableRunParams(BaseModel):
    """Request parameters for running an arbitrary executable."""

    executable_name: ExecutableName
    args: list[str] = []
    callback: HttpUrl | None = None
    is_async: bool = True


@router.post("/", response_model=ExecutableRunResponse, status_code=status.HTTP_201_CREATED)
async def run_executable_endpoint(params: ExecutableRunParams) -> ExecutableRunResponse:
    """Dispatch a task to run an arbitrary executable."""
    if params.is_async:
        job_id = run_executable_async(params.executable_name, params.args, params.callback)
        return ExecutableRunResponse(job_id=job_id)

    job_id = uuid4()
    result = await asyncio.to_thread(run_executable_sync, str(params.executable_name), params.args)
    return ExecutableRunResponse(job_id=job_id, result=result)
